import numpy as np
import scipy.sparse as sp


def cool_mean(data, partition):
    """Efficiently calculate the mean of all rows of a matrix M over a partition u. The number of classes in the
    partition is implicitly defined from the values of u.

    Parameters
    ----------
        data: Matrix of dimensions (n, f) with n data points of f features each.

        partition: Partition of the data points in the form of an (n, ) array with k different integer values.

    Returns
    -------
        group_mean: A (k, f) matrix with the vectors averaged over the k partition values.
    """
    s = data.shape[0]
    un, nf = np.unique(partition, return_counts=True)
    umat = sp.csr_matrix((np.ones(s, dtype='float32'), (np.arange(0, s), partition)), shape=(s, len(un)))
    return (umat.T @ data) / nf[..., np.newaxis]


def cool_max(arr, partition):
    """Efficiently calculate the max of all elements of an array arr of **positive** reals over a partition u.
    The number of classes in the partition is implicitly defined from the values of u.

    Parameters
    ----------
        arr: Array of dimensions (n, ).
        partition: Partition of the data points in the form of an (n, ) array with k different integer values.

    Returns
    -------
        group_max: A (k, ) array with the values maximized over the k partition values.
    """
    s = arr.size
    partition_size = np.max(partition) + 1
    umat = sp.csr_matrix((arr, (partition, np.arange(0, s))), shape=(partition_size, s))
    result = np.max(umat, axis=-1)
    return np.squeeze(result.toarray())


def cool_max_radius(data, partition):
    """Efficiently calculate the maximum norm of the rows of a matrix data over a partition u. The number of
    classes in the partition is implicitly defined from the values of u.

    Parameters
    ----------
        data: Matrix of dimensions (n, f) with n data points of f features each.

        partition: Partition of the data points in the form of an (n, ) array with k different integer values.

    Returns
    -------
        group_max_radius: A (k, ) array with the maximum vector norms over the k partition values.
    """
    norms = np.linalg.norm(data, axis=1)
    norm_maxes = cool_max(norms, partition)
    return norm_maxes


def cool_std(data, means, partition, epsilon=1e-12):
    """Efficiently calculate the standard deviation of all rows of a matrix data over a partition u. The means of
    each partition class are passed with the means matrix. The number of classes in the partition is implicitly
    defined from the values of u.

    Parameters
    ----------
        data: Matrix of dimensions (n, f) with n data points of f features each.
        means: Matrix of dimensions (n, f) with the means of the data per class. This implies that each mean is
            repeated over the vectors belonging to the same class.
        partition: Partition of the data points in the form of an (n, ) array with k different integer values.
        epsilon: Small constant to ensure that the standard deviation is not 0. This is specific to this codebase.

    Returns
    -------
        group_std: A (n, f) matrix with the standard deviation vectors over the k partition values. Rows belonging to
        the same partition class contain the same values.
    """
    return np.sqrt(cool_mean((data - means) ** 2, partition)) + epsilon


def cool_normalize(data, partition):
    """Efficiently normalize the rows of a matrix data over a partition u. The number of classes in the partition is
    implicitly defined from the values of u.

    Parameters
    ----------
        data: Matrix of dimensions (n, f) with n data points of f features each.

        partition: Partition of the data points in the form of an (n, ) array with k different integer values.

    Returns
    -------
        group_normalized: A (n, f) matrix of the original data normalized over the k partition classes.
    """
    means = cool_mean(data, partition)
    means_per_partition = means[partition]
    stds = cool_std(data, means_per_partition, partition)

    return (data - means_per_partition) / stds[partition], [means, stds]



import numpy as np

from hnne.cool_functions import cool_normalize


def norm_angle(data, theta, partition):
    rot = np.array([
        [np.cos(theta), np.sin(theta)],
        [-np.sin(theta), np.cos(theta)]
    ])

    data, norm1_params = cool_normalize(data, partition)

    rotated_data = np.dot(data, rot)
    rotated_data, norm2_params = cool_normalize(rotated_data, partition)

    return np.dot(rotated_data, np.linalg.inv(rot)), [rot, norm1_params, norm2_params]


def norm_angles(data, angles, partition_mapping):
    inflation_params = []
    for angle in angles:
        data, params = norm_angle(data, angle, partition_mapping)
        inflation_params.append(params)
    return data, inflation_params


def norm_angle_3d(data, alpha, beta, gamma, partition):
    r_x = np.array([
        [np.cos(alpha), -np.sin(alpha), 0],
        [np.sin(alpha), np.cos(alpha), 0],
        [0, 0, 1]
    ])
    r_y = np.array([
        [np.cos(beta), 0, np.sin(beta)],
        [0, 1, 0],
        [-np.sin(beta), 0, np.cos(beta)]
    ])
    r_z = np.array([
        [1, 0, 0],
        [0, np.cos(gamma), -np.sin(gamma)],
        [0, np.sin(gamma), np.cos(gamma)]
    ])

    rot = np.dot(r_x, np.dot(r_y, r_z))

    data, norm1_params = cool_normalize(data, partition)

    rotated_data = np.dot(data, rot)
    rotated_data, norm2_params = cool_normalize(rotated_data, partition)

    return np.dot(rotated_data, np.linalg.inv(rot)), [rot, norm1_params, norm2_params]


def norm_angles_3d(data, alphas, betas, gammas, partition_mapping):
    inflation_params = []
    for alpha, beta, gamma in zip(alphas, betas, gammas):
        data, params = norm_angle_3d(data, alpha, beta, gamma, partition_mapping)
        inflation_params.append(params)
    return data, inflation_params



import setuptools
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()


setuptools.setup(
    name="hnne",
    version="0.1.4",
    author="Marios Koulakis, Saquib Sarfraz",
    author_email="marios.koulakis@gmail.com",
    description="A fast hierarchical dimensionality reduction algorithm.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=['hnne'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: Free for non-commercial use",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="dimension dimensionality reduction t-sne umap hierarchical clustering finch",
    install_requires=[
        "scipy",
        "numpy==1.20",
        "sklearn",
        "tqdm",
        "pynndescent",
        "typer",
        "pandas",
        "cython",
    ],
    test_suite="pytest",
    tests_require=["pytest"],
)
